/*
    Created from:
        Messages = <INPUTFILENAME>
        Template = <TEMPLATEFILENAME>
        Language = <LANGUAGEFILENAME>

                     AUTOGENERATED FILE, DO NOT EDIT

*/
#ifndef <MSGFULLNAME>Code_H__
#define <MSGFULLNAME>Code_H__

#define <MSGFULLNAME>_MSG_ID <MSGID>
#define <MSGFULLNAME>_MSG_SIZE <MSGSIZE>
<ENUMERATIONS>
<FIELDINFOS>
<ACCESSORS>
INLINE void <MSGFULLNAME>_Init(uint8_t* m_data)
{
    <INIT_CODE>
}

#endif
