%{
    Created <DATE> from:
        Messages = <INPUTFILENAME>
        Template = <TEMPLATEFILENAME>
        Language = <LANGUAGEFILENAME>

                     AUTOGENERATED FILE, DO NOT EDIT

%}

classdef <MSGSHORTNAME>

    properties (Constant)
        MSG_ID = <MSGID>;
        MSG_SIZE = <MSGSIZE>;
        <ENUMERATIONS>
    end
    properties (Hidden)
        % the buffer of data
        m_data;
    end
    properties  (Dependent)
            <DECLARATIONS>
    end
    methods
        function obj = <MSGSHORTNAME>(data)
            if nargin == 0
                % create a new message (allocates a buffer for it)
                obj.m_data = zeros(obj.MSG_SIZE, 1, 'uint8');
                <INIT_CODE>
            else
                % create a message based on a data buffer
                obj.m_data = data;
            end
        end
    end
    methods
        <ACCESSORS>
    end
end