/*
    <OUTPUTFILENAME>
    Created <DATE> from:
        Messages = <INPUTFILENAME>
        Template = <TEMPLATEFILENAME>
        Language = <LANGUAGEFILENAME>

                     AUTOGENERATED FILE, DO NOT EDIT

*/
import Foundation

public class <MSGNAME>
{
    public static let MSG_SIZE = <MSGSIZE>;
    public static let MSG_NAME = "<MSGDESCRIPTOR>";
    var m_data: NSMutableData;
    
    init()
    {
        m_data = NSMutableData(length: <MSGNAME>.MSG_SIZE)!;
        Initialize();
    }
    init(buffer: NSMutableData)
    {
        m_data = buffer;
    }
    func Initialize()
    {
        <INIT_CODE>
    }

    public func SetMessageID(_ id: UInt32)
    {
        <SETMSGID>;
    };
    public func GetMessageID() -> UInt32
    {
        <GETMSGID>
    };

    <ENUMERATIONS>
    <ACCESSORS>
};
