/*
    <OUTPUTFILENAME>
    Created <DATE> from:
        Messages = <INPUTFILENAME>
        Template = <TEMPLATEFILENAME>
        Language = <LANGUAGEFILENAME>

                     AUTOGENERATED FILE, DO NOT EDIT

*/
package <MESSAGE_PACKAGE>

import java.nio.ByteBuffer

class <MSGSHORTNAME>: msgplugin.Message {
    companion object {
        val MSG_ID = <MSGID>
        val MSG_SIZE = <MSGSIZE>
        val MSG_NAME = "<MSGDESCRIPTOR>"
    }

    constructor(): super(MSG_SIZE) {
        setMessageId(MSG_ID)
        initialize()
    }

    constructor(buffer: ByteBuffer): super(buffer) {
        initialize()
    }

    fun initialize() {
        <INIT_CODE>
    }

    <ENUMERATIONS>
    <FIELDINFOS>
    <ACCESSORS>
}
