/*
    Created from:
        Messages = <INPUTFILENAME>
        Template = <TEMPLATEFILENAME>
        Language = <LANGUAGEFILENAME>

                     AUTOGENERATED FILE, DO NOT EDIT

*/
#ifndef <MSGNAME>Code_H__
#define <MSGNAME>Code_H__

#define <MSGNAME>_MSG_SIZE <MSGSIZE>
<ENUMERATIONS>
<FIELDINFOS>
<ACCESSORS>
INLINE void <MSGNAME>_SetMessageID(uint8_t* m_data, uint32_t id)
{
    <SETMSGID>;
}
INLINE uint32_t <MSGNAME>_GetMessageID(uint8_t* m_data)
{
    return <GETMSGID>;
}
INLINE void <MSGNAME>_Init(uint8_t* m_data)
{
    <INIT_CODE>
}

#endif
